import { InputNode } from '../../models/types/inputs/input-node';
import { UseContext } from '../../models/decorators/context/form-context';
import 'reflect-metadata';
import { InputProcessorService } from '../input-processor/input-processor.service';
import * as i0 from "@angular/core";
export declare class FormEntityProcessorService {
    private inputProcessor;
    constructor(inputProcessor: InputProcessorService);
    /**
     *Builds tree like structure from form entities
     *
     * @param formEntity class annotated with `FormEntity`
     * @param context `string` for proper usage please go to docs
     * @returns `InputNode` tree of input nodes
     */
    process(formEntity: any, context?: UseContext): InputNode;
    /**
     * Used internally to build input nodes
     *
     * @param entity class annotated with `FormEntity`
     * @param context `string` for proper usage please go to docs
     * @param parentProperties `Map<string, any>`
     * @returns `InputNode`
     */
    private parse;
    /**
     *Returns attributes within given context
     *
     * @param entity
     * @param context
     * @returns `string[]` attributes within context
     */
    private processUseContext;
    /**
     * Used internally to prepare nested form entity and its properties
     * to be used with `createNode`
     *
     * @param entity class with `@FormEntity`
     * @param context `string`
     * @param key `string`
     * @param formProperties parent form properties `Map<string, any>` current impl overrides nested form properties with its parent this may change in the future
     * @returns `InputNode` tree like that holds nodes
     */
    private processNestedForm;
    /**
     * Return qupdate strategy form mrta data if not set default `change`
     *
     * @param formProperties
     * @returns
     */
    private processUpdateStrategy;
    /**
     * Applies cross validation to form
     * @param entity
     * @param formNode
     */
    private processCrossValidation;
    /**
     * Handles `enableFn` meta attribute
     * @param root `InputNode`
     */
    private processEnableFn;
    /**
     * Used to apply `enableFn` after setting from value manually and silent change events
     * so keep it consistent
     *
     * @param root `InputNode`
     */
    applyEnableFn(root: InputNode): void;
    /**
     * Checks each field recursively to enable or disable it
     * if `enableFn` is defined on it
     *
     * @param root `InputNode`
     * @param formValue `Object`
     */
    private applyIfDefined;
    /**
     * Sorts inputs ASC based on `order` `property`
     * @param inputNode `InputNode`
     *
     */
    private processOrder;
    private isNotComposite;
    private isComposite;
    static ɵfac: i0.ɵɵFactoryDeclaration<FormEntityProcessorService, never>;
    static ɵprov: i0.ɵɵInjectableDeclaration<FormEntityProcessorService>;
}
